@@ -1,7 +1,7 @@
 # +--------------------------------------------------------------------------+
 # |  Licensed Materials - Property of IBM                                    |
 # |                                                                          |
-# | (C) Copyright IBM Corporation 2008, 2013.                                |
+# | (C) Copyright IBM Corporation 2008, 2014.                                |
 # +--------------------------------------------------------------------------+
 # | This module complies with SQLAlchemy 0.8 and is                          |
 # | Licensed under the Apache License, Version 2.0 (the "License");          |
@@ -25,13 +25,14 @@ from sqlalchemy import schema as sa_schema
 from sqlalchemy import util
 from sqlalchemy.sql import compiler
 from sqlalchemy.engine import default
-
+from sqlalchemy import __version__ as SA_Version
 from . import reflection as ibm_reflection
 
 from sqlalchemy.types import BLOB, CHAR, CLOB, DATE, DATETIME, INTEGER,\
     SMALLINT, BIGINT, DECIMAL, NUMERIC, REAL, TIME, TIMESTAMP,\
-    VARCHAR
+    VARCHAR, FLOAT
 
+SA_Version = [int(ver_token) for ver_token in SA_Version.split('.')[0:2]]
 
 # as documented from:
 # http://publib.boulder.ibm.com/infocenter/db2luw/v9/index.jsp?topic=/com.ibm.db2.udb.doc/admin/r0001095.htm
@@ -170,7 +171,7 @@ class XML(sa_types.Text):
 
 colspecs = {
     sa_types.Boolean: _IBM_Boolean,
-    sa_types.Date: _IBM_Date,
+    sa_types.Date: _IBM_Date
 # really ?
 #    sa_types.Unicode: DB2VARGRAPHIC
 }
@@ -189,6 +190,7 @@ ischema_names = {
     'NUMERIC': NUMERIC,
     'REAL': REAL,
     'DOUBLE': DOUBLE,
+    'FLOAT': FLOAT,
     'TIME': TIME,
     'TIMESTAMP': TIMESTAMP,
     'VARCHAR': VARCHAR,
@@ -229,6 +231,9 @@ class DB2TypeCompiler(compiler.GenericTypeCompiler):
         return "FLOAT" if type_.precision is None else \
                 "FLOAT(%(precision)s)" % {'precision': type_.precision}
 
+    def visit_DOUBLE(self, type_):
+        return "DOUBLE"
+
     def visit_XML(self, type_):
         return "XML"
 
@@ -315,6 +320,14 @@ class DB2Compiler(compiler.SQLCompiler):
     def visit_now_func(self, fn, **kw):
         return "CURRENT_TIMESTAMP"
 
+    def for_update_clause(self, select):
+        if select.for_update == True:
+            return ' WITH RS USE AND KEEP UPDATE LOCKS'
+        elif select.for_update == 'read':
+            return ' WITH RS USE AND KEEP SHARE LOCKS'
+        else:
+            return ''
+
     def visit_mod_binary(self, binary, operator, **kw):
         return "mod(%s, %s)" % (self.process(binary.left),
                                                 self.process(binary.right))
@@ -336,12 +349,14 @@ class DB2Compiler(compiler.SQLCompiler):
     def default_from(self):
         # DB2 uses SYSIBM.SYSDUMMY1 table for row count
         return  " FROM SYSIBM.SYSDUMMY1"
-    
+
     def visit_function(self, func, result_map=None, **kwargs):
         if func.name.upper() == "AVG":
             return "AVG(DOUBLE(%s))" % (self.function_argspec(func, **kwargs))
+        elif func.name.upper() == "CHAR_LENGTH":
+            return "CHAR_LENGTH(%s, %s)" % (self.function_argspec(func, **kwargs), 'OCTETS')
         else:
-            return compiler.SQLCompiler.visit_function(self, func, **kwargs)        
+            return compiler.SQLCompiler.visit_function(self, func, **kwargs)
     # TODO: this is wrong but need to know what DB2 is expecting here
     #    if func.name.upper() == "LENGTH":
     #        return "LENGTH('%s')" % func.compile().params[func.name + '_1']
@@ -379,16 +394,16 @@ class DB2Compiler(compiler.SQLCompiler):
              self.process(join.right, asfrom=True, **kwargs),
              " ON ",
              self.process(join.onclause, **kwargs)))
-    
+
     def visit_savepoint(self, savepoint_stmt):
         return "SAVEPOINT %(sid)s ON ROLLBACK RETAIN CURSORS" % {'sid':self.preparer.format_savepoint(savepoint_stmt)}
 
     def visit_rollback_to_savepoint(self, savepoint_stmt):
         return 'ROLLBACK TO SAVEPOINT %(sid)s'% {'sid':self.preparer.format_savepoint(savepoint_stmt)}
-       
+
     def visit_release_savepoint(self, savepoint_stmt):
         return 'RELEASE TO SAVEPOINT %(sid)s'% {'sid':self.preparer.format_savepoint(savepoint_stmt)}
-        
+
 class DB2DDLCompiler(compiler.DDLCompiler):
 
     def get_column_specification(self, column, **kw):
@@ -425,13 +440,6 @@ class DB2DDLCompiler(compiler.DDLCompiler):
 
         return text
 
-    def visit_drop_index(self, drop, **kw):
-        return "\nDROP INDEX %s" % (
-                        self.preparer.quote(
-                                    self._index_identifier(drop.element.name),
-                                    drop.element.quote)
-                        )
-
     def visit_drop_constraint(self, drop, **kw):
         constraint = drop.element
         if isinstance(constraint, sa_schema.ForeignKeyConstraint):
@@ -449,7 +457,55 @@ class DB2DDLCompiler(compiler.DDLCompiler):
         return "ALTER TABLE %s DROP %s%s" % \
                                 (self.preparer.format_table(constraint.table),
                                 qual, const)
+                                
+    def create_table_constraints(self, table, **kw):
+        if (self.dialect.dbms_name.find('DB2/') != -1) and ([int(ver_token) for ver_token in self.dialect.dbms_ver.split('.')[0:2]] >= [10, 5]):
+            for constraint in table._sorted_constraints:
+                if isinstance(constraint, sa_schema.UniqueConstraint):
+                    for column in constraint.columns._all_columns:
+                        if column.nullable:
+                            constraint.use_alter = True
+                            break
+                    if hasattr(constraint, 'use_alter') and constraint.use_alter:
+                        if not constraint.name:
+                            index_name = "%s_%s_%s" % ('ukey', self.preparer.format_table(constraint.table), '_'.join(column.name for column in constraint.columns._all_columns))
+                        else:
+                            index_name = constraint.name
+                        index = sa_schema.Index(index_name, *(column for column in constraint.columns._all_columns))
+                        index.unique = True
+                        index.exclude_nulls = True
+        result = super( DB2DDLCompiler, self ).create_table_constraints(table, **kw)
+        return result
+    
+    def visit_create_index(self, create, include_schema=True, include_table_schema=True):
+        if SA_Version < [0, 8]:
+            sql = super( DB2DDLCompiler, self ).visit_create_index(create)
+        else:
+            sql = super( DB2DDLCompiler, self ).visit_create_index(create, include_schema, include_table_schema)
+        if getattr(create.element, 'exclude_nulls', None):
+            sql += ' EXCLUDE NULL KEYS'
+        return sql
 
+    def visit_add_constraint(self, create):
+        if (self.dialect.dbms_name.find('DB2/') != -1) and ([int(ver_token) for ver_token in self.dialect.dbms_ver.split('.')[0:2]] >= [10, 5]):
+            if isinstance(create.element, sa_schema.UniqueConstraint):
+                for column in create.element.columns._all_columns:
+                    if column.nullable:
+                        create.element.exclude_nulls = True
+                        break
+                if getattr(create.element, 'exclude_nulls', None):
+                    if not create.element.name:
+                        index_name = "%s_%s_%s" % ('uk_index', self.preparer.format_table(create.element.table), '_'.join(column.name for column in create.element.columns._all_columns))
+                    else:
+                        index_name = create.element.name
+                    index = sa_schema.Index(index_name, *(column for column in create.element.columns._all_columns))
+                    index.unique = True
+                    index.exclude_nulls = True
+                    sql = self.visit_create_index(sa_schema.CreateIndex(index)) 
+                    return sql
+        sql = super( DB2DDLCompiler, self ).visit_add_constraint(create)
+        return sql
+    
 class DB2IdentifierPreparer(compiler.IdentifierPreparer):
 
     reserved_words = RESERVED_WORDS
@@ -461,13 +517,13 @@ class DB2ExecutionContext(default.DefaultExecutionContext):
         return self._execute_scalar("SELECT NEXTVAL FOR " +
                     self.dialect.identifier_preparer.format_sequence(seq) +
                     " FROM SYSIBM.SYSDUMMY1", type_)
-                    
+
 class _SelectLastRowIDMixin(object):
     _select_lastrowid = False
     _lastrowid = None
 
 
-    def getlastrowid(self):
+    def get_lastrowid(self):
         return self._lastrowid
 
     def pre_exec(self):
@@ -496,14 +552,17 @@ class DB2Dialect(default.DefaultDialect):
     name = 'ibm_db_sa'
     max_identifier_length = 128
     encoding = 'utf-8'
-    default_paramstyle = 'named'
+    default_paramstyle = 'qmark'
     colspecs = colspecs
     ischema_names = ischema_names
     supports_char_length = False
+    supports_unicode_statements = False
+    supports_unicode_binds = False
+    returns_unicode_strings = False
     postfetch_lastrowid = True
     supports_sane_rowcount = True
     supports_sane_multi_rowcount = True
-    supports_native_decimal = True
+    supports_native_decimal = False
     preexecute_sequences = False
     supports_alter = True
     supports_sequences = True
@@ -515,7 +574,7 @@ class DB2Dialect(default.DefaultDialect):
     supports_empty_insert = False
 
     two_phase_transactions = False
-    savepoints =  True 
+    savepoints =  True
 
     statement_compiler = DB2Compiler
     ddl_compiler = DB2DDLCompiler
@@ -532,7 +591,11 @@ class DB2Dialect(default.DefaultDialect):
 
     # reflection: these all defer to an BaseDB2Reflector
     # object which selects between DB2 and AS/400 schemas
-
+    def initialize(self, connection):
+        super(DB2Dialect, self).initialize(connection)
+        self.dbms_ver = connection.connection.dbms_ver
+        self.dbms_name = connection.connection.dbms_name
+        
     def normalize_name(self, name):
         return self._reflector.normalize_name(name)
 
