@@ -29,6 +29,8 @@ class CoerceUnicode(sa_types.TypeDecorator):
     impl = sa_types.Unicode
 
     def process_bind_param(self, value, dialect):
+        if isinstance(value, str):
+            value = value.decode(dialect.encoding)
         return value
 
 class BaseReflector(object):
@@ -40,7 +42,7 @@ class BaseReflector(object):
     def normalize_name(self, name):
         if isinstance(name, str):
             name = name.lower()
-        elif name != None:
+        if name != None:
             return name.lower() if name.upper() == name and \
                not self.identifier_preparer._requires_quotes(name.lower()) \
                else name
@@ -64,6 +66,8 @@ class BaseReflector(object):
                     'SELECT CURRENT_SCHEMA FROM SYSIBM.SYSDUMMY1').scalar()
         if isinstance(default_schema_name, str):
             default_schema_name = default_schema_name.strip()
+        elif isinstance(default_schema_name, str):
+            default_schema_name = default_schema_name.strip().__str__()
         return self.normalize_name(default_schema_name)
 
     @property
@@ -121,6 +125,8 @@ class DB2Reflector(BaseReflector):
       Column("SCALE", sa_types.Integer, key="scale"),
       Column("DEFAULT", CoerceUnicode, key="defaultval"),
       Column("NULLS", CoerceUnicode, key="nullable"),
+      Column("IDENTITY", CoerceUnicode, key="identity"),
+      Column("GENERATED", CoerceUnicode, key="generated"),
       schema="SYSCAT")
 
     sys_views = Table("VIEWS", ischema,
@@ -185,7 +191,7 @@ class DB2Reflector(BaseReflector):
         query = sql.select([self.sys_views.c.viewname]).\
             where(self.sys_views.c.viewschema == current_schema).\
             order_by(self.sys_views.c.viewname)
-            
+
         return [self.normalize_name(r[0]) for r in connection.execute(query)]
 
     @reflection.cache
@@ -196,7 +202,7 @@ class DB2Reflector(BaseReflector):
         query = sql.select([self.sys_views.c.text]).\
             where(self.sys_views.c.viewschema == current_schema).\
             where(self.sys_views.c.viewname == viewname)
-            
+
         return connection.execute(query).scalar()
 
     @reflection.cache
@@ -207,7 +213,8 @@ class DB2Reflector(BaseReflector):
 
         query = sql.select([syscols.c.colname, syscols.c.typename,
                             syscols.c.defaultval, syscols.c.nullable,
-                            syscols.c.length, syscols.c.scale],
+                            syscols.c.length, syscols.c.scale,
+                            syscols.c.identity, syscols.c.generated],
               sql.and_(
                   syscols.c.tabschema == current_schema,
                   syscols.c.tabname == table_name
@@ -235,6 +242,7 @@ class DB2Reflector(BaseReflector):
                     'type': coltype,
                     'nullable': r[3] == 'Y',
                     'default': r[2] or None,
+                    'autoincrement': (r[6] == 'Y') and (r[7] != ' '),
                 })
         return sa_columns
 
@@ -260,7 +268,9 @@ class DB2Reflector(BaseReflector):
 
     @reflection.cache
     def get_foreign_keys(self, connection, table_name, schema=None, **kw):
-        current_schema = self.denormalize_name(schema or self.default_schema_name)
+        default_schema = self.default_schema_name
+        current_schema = self.denormalize_name(schema or default_schema)
+        default_schema = self.normalize_name(default_schema)
         table_name = self.denormalize_name(table_name)
         sysfkeys = self.sys_foreignkeys
         query = sql.select([sysfkeys.c.fkname, sysfkeys.c.fktabschema, \
@@ -282,15 +292,15 @@ class DB2Reflector(BaseReflector):
                 # if no schema specified and referred schema here is the
                 # default, then set to None
                 if schema is None and \
-                    referred_schema == self.default_schema_name:
+                    referred_schema == default_schema:
                     referred_schema = None
 
                 fschema[r[0]] = {
                     'name': self.normalize_name(r[0]),
-                  'constrained_columns': [self.normalize_name(r[3])],
-                  'referred_schema': referred_schema,
-                  'referred_table': self.normalize_name(r[6]),
-                  'referred_columns': [self.normalize_name(r[7])]}
+                    'constrained_columns': [self.normalize_name(r[3])],
+                    'referred_schema': referred_schema,
+                    'referred_table': self.normalize_name(r[6]),
+                    'referred_columns': [self.normalize_name(r[7])]}
             else:
                 fschema[r[0]]['constrained_columns'].append(self.normalize_name(r[3]))
                 fschema[r[0]]['referred_columns'].append(self.normalize_name(r[7]))
@@ -364,6 +374,8 @@ class AS400Reflector(BaseReflector):
       Column("IS_NULLABLE", sa_types.Integer, key="nullable"),
       Column("COLUMN_DEFAULT", CoerceUnicode, key="defaultval"),
       Column("HAS_DEFAULT", CoerceUnicode, key="hasdef"),
+      Column("IS_IDENTITY", CoerceUnicode, key="isid"),
+      Column("IDENTITY_GENERATION", CoerceUnicode, key="idgenerate"),
       schema="QSYS2")
 
     sys_indexes = Table("SYSINDEXES", ischema,
@@ -446,13 +458,12 @@ class AS400Reflector(BaseReflector):
     # Retrieves a list of table names for a given schema
     @reflection.cache
     def get_table_names(self, connection, schema=None, **kw):
-        current_schema = self.denormalize_name(
-                            schema or self.default_schema_name)
+        current_schema = self.denormalize_name(schema or self.default_schema_name)
         systbl = self.sys_tables
-        query = sql.select([systbl.c.tabname],
-                systbl.c.tabschema == current_schema,
-                order_by=[systbl.c.tabname]
-            )
+        query = sql.select([systbl.c.tabname]).\
+                where(systbl.c.tabtype == 'T').\
+                where(systbl.c.tabschema == current_schema).\
+                order_by(systbl.c.tabname)
         return [self.normalize_name(r[0]) for r in connection.execute(query)]
 
     @reflection.cache
@@ -480,21 +491,20 @@ class AS400Reflector(BaseReflector):
 
     @reflection.cache
     def get_columns(self, connection, table_name, schema=None, **kw):
-        current_schema = self.denormalize_name(
-                                schema or self.default_schema_name)
+        current_schema = self.denormalize_name(schema or self.default_schema_name)
         table_name = self.denormalize_name(table_name)
         syscols = self.sys_columns
 
         query = sql.select([syscols.c.colname,
                                 syscols.c.typename,
                                 syscols.c.defaultval, syscols.c.nullable,
-                                syscols.c.length, syscols.c.scale],
+                                syscols.c.length, syscols.c.scale,
+                                syscols.c.isid, syscols.c.idgenerate],
                     sql.and_(
                             syscols.c.tabschema == current_schema,
                             syscols.c.tabname == table_name
                         ),
-                    order_by=[syscols.c.tabschema, syscols.c.tabname,
-                                    syscols.c.colname, syscols.c.colno]
+                    order_by=[syscols.c.colno]
                 )
         sa_columns = []
         for r in connection.execute(query):
@@ -517,7 +527,7 @@ class AS400Reflector(BaseReflector):
                     'type': coltype,
                     'nullable': r[3] == 'Y',
                     'default': r[2],
-                    'autoincrement': r[2] is None,
+                    'autoincrement': (r[6] == 'YES') and (r[7] != None),
                 })
         return sa_columns
 
@@ -543,8 +553,9 @@ class AS400Reflector(BaseReflector):
 
     @reflection.cache
     def get_foreign_keys(self, connection, table_name, schema=None, **kw):
-        current_schema = self.denormalize_name(
-                                    schema or self.default_schema_name)
+        default_schema = self.default_schema_name
+        current_schema = self.denormalize_name(schema or default_schema)
+        default_schema = self.normalize_name(default_schema)
         table_name = self.denormalize_name(table_name)
         sysfkeys = self.sys_foreignkeys
         query = sql.select([sysfkeys.c.fkname, sysfkeys.c.fktabschema, \
@@ -560,16 +571,22 @@ class AS400Reflector(BaseReflector):
         fschema = {}
         for r in connection.execute(query):
             if r[0] not in fschema:
+                referred_schema = self.normalize_name(r[5])
+
+                # if no schema specified and referred schema here is the
+                # default, then set to None
+                if schema is None and \
+                    referred_schema == default_schema:
+                    referred_schema = None
+
                 fschema[r[0]] = {'name': self.normalize_name(r[0]),
                             'constrained_columns': [self.normalize_name(r[3])],
-                            'referred_schema': self.normalize_name(r[5]),
+                            'referred_schema': referred_schema,
                             'referred_table': self.normalize_name(r[6]),
                             'referred_columns': [self.normalize_name(r[7])]}
             else:
-                fschema[r[0]]['constrained_columns'].append(
-                                                    self.normalize_name(r[3]))
-                fschema[r[0]]['referred_columns'].append(
-                                                    self.normalize_name(r[7]))
+                fschema[r[0]]['constrained_columns'].append(self.normalize_name(r[3]))
+                fschema[r[0]]['referred_columns'].append(self.normalize_name(r[7]))
         return [value for key, value in fschema.items()]
 
     # Retrieves a list of index names for a given schema
